'''

Copyright (C) 2019 Vanessa Sochat.

This Source Code Form is subject to the terms of the
Mozilla Public License, v. 2.0. If a copy of the MPL was not distributed
with this file, You can obtain one at http://mozilla.org/MPL/2.0/.

'''

from watchme.command import get_watchers
from watchme import get_watcher
from watchme.tasks.decorators import TerminalRunner
import json

def main(args, extra):
    '''monitor a task (from the command line), meaning wrapping it with
       multiprocessing, getting the id, and returning a result (command line 
       or written to file)
    '''  
    # The first argument will either be part of a command, or a watcher
    watcher = args.watcher

    # The entire user command is the extra arguments
    command = extra

    # If the user provides a watcher, we are saving to it
    if watcher not in get_watchers(args.base, quiet=True):       
        command = [watcher] + command
        watcher = None
    else:
        watcher = get_watcher(watcher, base=args.base, create=False)

    command = ' '.join(command)
    runner = TerminalRunner(command, 
                            skip=args.skip,
                            include=args.include,
                            only=args.only,
                            seconds=args.seconds)
    runner.run()
    timepoints = runner.wait(args.func)

    # The output folder depends on the watcher func
    prefix = "decorator-psutils"
    if args.func == "gpu_task":
        prefix = "decorator-gpu"

    # If we don't have a watcher, print to terminal
    if watcher is None or args.test is True:
        print(json.dumps(timepoints))

    # Otherwise save to watcher task folder
    else:
        name = args.name
        if name is None:
            name = command.replace(' ', '-')
        name = '%s-%s' % (prefix, name)
        watcher.finish_runs({name: timepoints})



#!/usr/bin/python

# Copyright (C) 2019 Vanessa Sochat.

# This Source Code Form is subject to the terms of the
# Mozilla Public License, v. 2.0. If a copy of the MPL was not distributed
# with this file, You can obtain one at http://mozilla.org/MPL/2.0/.

from watchme import get_watcher
from watchme.command import create_watcher
import unittest
import tempfile
import shutil


print("####################################################### test_decorators")


class TestDecorators(unittest.TestCase):

    def setUp(self):
        self.base = tempfile.mkdtemp()
        self.repo = create_watcher('pancakes', base=self.base)
        self.cli = get_watcher('pancakes', base=self.base)
        
    def tearDown(self):
        shutil.rmtree(self.base)

    def test_psutils_monitor(self):
        '''test creation function, and basic watcher config'''
        print("Testing psutilsc.decorators.TerminalRunner")
        from watchme.tasks.decorators import TerminalRunner
        runner = TerminalRunner('sleep 2')
        runner.run()
        timepoints = runner.wait("monitor_pid_task")
        self.assertTrue(len(timepoints)==1)

if __name__ == '__main__':
    unittest.main()



'''

Copyright (C) 2019 Vanessa Sochat.

This Source Code Form is subject to the terms of the
Mozilla Public License, v. 2.0. If a copy of the MPL was not distributed
with this file, You can obtain one at http://mozilla.org/MPL/2.0/.

'''

from watchme.logger import bot
from watchme.utils import (
    write_file,
    write_json
)

import shutil
import os

class TaskBase(object):

    def __init__(self, name, params=None, **kwargs):

        # If params not defined, create as dict
        if params is None:
            params = {}

        # Ensure subclass was created correctly
        for req in ['required_params', 'type', 'export_func']:
            if not hasattr(self, req):
                bot.exit('A Task must have a %s function or attribute.' % req)

        self.name = name
        self.valid = False
        self.params = {}
        self.set_params(params)
        self.validate()

    def get_type(self):
        '''get the watcher type.
        '''
        return self.type


# Identification

    def __repr__(self):
        return "[task|%s]" %self.name

    def __str__(self):
        return "[task|%s]" %self.name


# Parameters

    def set_params(self, params):
        '''iterate through parameters, set into dictionary.

           Parameters
           ==========
           params: a list of key@value pairs to set.
        '''
        for key,value in params.items():
            key = key.lower()
            self.params[key] = value


    def export_params(self, active="true"):
        '''export parameters, meaning returning a dictionary of the task
           parameters plus the addition of the task type and active status.
        '''
        params = self.params.copy()
        params['active'] = active
        params['type'] = self.type
        return params


# Validation

    def validate(self):
        '''validate the parameters set for the Task. Exit if there are any
           errors. Ensure required parameters are defined, and have correct
           values.
        '''
        self.valid = True

        for param in self.required_params:
            if param not in self.params:
                bot.error('Missing required parameter: %s' % param)
                self.valid = False

        # Call subclass validation function
        self._validate()

    def _validate(self):
        '''validation function intended to be implemented by subclass.
        '''
        return


# Run Single Task

    def run(self):
        '''run an isolated task, meaning no update or communication with
           the watcher. This will return the raw result.
        '''
        params = self.export_params()
        func = self.export_func()
        if func is not None:
            return func(**params)
        bot.error('Cannot find function.')


# Save Entrypoint

    def write_results(self, result, repo):
        '''an entrypoint function for a general task. By default, we parse
           results based on the result type. Any particular subclass of the
           TaskBase can modify or extend these functions.

           Parameters
           ==========
           result: the result object to parse
           repo: the repo base (watcher.repo)
        '''
        files = []

        # Case 1. The result is a list
        if isinstance(result, list):

            # Get rid of Nones, if the user accidentally added
            result = [r for r in result if r]

            if len(result) == 0:
                bot.error('%s returned empty list of results.' % self.name)

            # multiple jsons save specified, regardless
            elif self.params.get('save_as') == 'jsons':
                bot.debug('Saving single list as multiple json...')
                files += self._save_json_list(result, repo)

            # json output is specified by the user or we find dict results
            elif self.params.get('save_as') == 'json' or isinstance(result[0], dict):
                bot.debug('Saving single list as one json...')
                files.append(self._save_json(result, repo))

            # Otherwise, sniff for list of paths
            elif os.path.exists(result[0]):
                bot.debug('Found list of paths...')
                files += self._save_files_list(result, repo)

            # Finally, assume just writing text to file
            else:
                bot.debug('Saving content from list to file...')
                files += self._save_text_list(result, repo)

        # Case 2. The result is a string
        elif isinstance(result, str):
            files = self._save_str_result(files, result, repo)

        # Case 3. The result is a dictionary
        elif isinstance(result, dict):
            files.append(self._save_json(result,repo))

        elif result is None:
            bot.error('Result for task %s is None' % self.name)

        elif hasattr(self, '_write_results'):
            return self._write_results(result)

        # If it's unicode, try encoding, and then fail (repetitive)
        else:
            try:
                result = result.encode('utf-8')
                files = self._save_str_result(files, result, repo)

            except:
                bot.error('Unsupported result format %s' % type(result))

        # Get rid of None results (don't check excessively for None above)
        files = [f for f in files if f]
        return files


# Saving

    def _save_str_result(self, files, result, repo):
        '''a helper function to return a list of files with added string
           results. We do this twice in the saving function.

           Parameters
           ==========
           files: the list of files to save (returned at the end)
           result: the result object to parse, should be string
           repo: the repo to write it to.
        '''
        # if it's a path to a file, just save to repository
        if os.path.exists(result):
            files.append(self._save_file(result, repo))

        # Otherwise, it's a string that needs to be saved to file
        else:
            files.append(self._save_text(result, repo))

        return files


    def _save_list(self, results, repo, func, file_name):
        '''general function to perform saving a list of content to json,
           text, or moving paths. Used by save_file_list, save_json_list,
           and save_text_list.

           Parameters
           ==========
           results: list of results, assumed to be the correct type
           repo: the repository base with the task folder
           func: the function to call.
           file_name: the base file name to use.
        '''
        file_name, ext = os.path.splitext(file_name)
        files = []

        for r, result in enumerate(results):
            filename = "%s-%s%s" %(file_name, str(r), ext)
            saved = func(result, repo, filename)
            files.append(saved)

        return files


    def _save_text(self, result, repo, file_name=None):
        '''save a general text object to file.
 
           Parameters
           ==========
           result: the result object to save, should be path to a file
           repo: the repository base with the task folder
        '''
        file_name = self.params.get('file_name', file_name) or 'result.txt'
        task_folder = os.path.join(repo, self.name)
        destination = os.path.join(task_folder, file_name)
        write_file(destination, result)
        return destination


    def _save_file(self, result, repo, file_name=None):
        '''for a result that exists, move the file to final destination.
 
           Parameters
           ==========
           result: the result object to save, should be path to a file
           repo: the repository base with the task folder
        '''
        if os.path.exists(result):
            name = os.path.basename(result)
            file_name = self.params.get('file_name', file_name) or name
            task_folder = os.path.join(repo, self.name)

            # HOME/.watchme/watcher/<task>/<result>
            destination = os.path.join(task_folder, file_name)
            shutil.move(result, destination)

            # <task>/<result>     
            return os.path.join(name, file_name)

        bot.warning('%s does not exist.' % result)

        
    def _save_json(self, result, repo, file_name=None):
        '''for a result that is a dictionary or list, save as json
 
           Parameters
           ==========
           result: the result object to save, should be dict or list
        '''
        file_name = self.params.get('file_name', file_name) or 'result.json'
        destination = os.path.join(repo, self.name, file_name)
        write_json(result, destination)
        return destination


    def _save_files_list(self, results, repo):
        '''If the user provides already existing files, we simply move them
           into the task folder in the repository.
 
           Parameters
           ==========
           results: the results to save, should a list of files
        '''
        files = []
        for result in results:
            filename = os.path.basename(result)
            filename = os.path.join(repo, self.name, filename)
            shutil.move(result, filename) 
            files.append(filename)

        return files

    def _save_json_list(self, results, repo):
        '''A wrapper around save json for a list, handles file naming.
 
           Parameters
           ==========
           results: the results to save, should a list
        '''
        file_name = self.params.get('file_name', 'result.json')
        return self._save_list(results, repo, self._save_json, file_name)


    def _save_file_list(self, results, repo):
        '''for a list of results that exist, move the files to 
           final destination.
 
           Parameters
           ==========
           results: list of paths to a file, those not existing are skipped
           repo: the repository base with the task folder
        '''
        file_name = self.params.get('file_name', os.path.basename(results))
        return self._save_list(results, repo, self._save_file, file_name)


    def _save_text_list(self, results, repo):
        '''for a list of general text results, write them to output files.
 
           Parameters
           ==========
           results: list of string results to write to file
           repo: the repository base with the task folder
        '''
        file_name = self.params.get('file_name', 'result.txt')
        return self._save_list(results, repo, self._save_text, file_name)



'''

Copyright (C) 2019 Vanessa Sochat.

This Source Code Form is subject to the terms of the
Mozilla Public License, v. 2.0. If a copy of the MPL was not distributed
with this file, You can obtain one at http://mozilla.org/MPL/2.0/.

'''

from watchme import get_watcher

def main(args, extra):
    '''protect or freeze a watcher, or disable.
    '''    
    # Required - will print help if not provided
    name = args.watcher[0]
    watcher = get_watcher(name, base=args.base, create=False)

    if args.action in ['on', 'off']:
        watcher.protect(args.action)
    elif args.action == "freeze":
        watcher.freeze()
    elif args.action == "unfreeze":
        watcher.unfreeze()



'''

Copyright (C) 2019 Vanessa Sochat.

This Source Code Form is subject to the terms of the
Mozilla Public License, v. 2.0. If a copy of the MPL was not distributed
with this file, You can obtain one at http://mozilla.org/MPL/2.0/.

'''

from watchme.utils import get_watchme_env
from .pynvml import (
    nvmlInit, 
    nvmlShutdown
)
from watchme.watchers.gpu import pynvml

def gpu_task(**kwargs):
    '''Get variables about the gpu of the host. No parameters are required.
       We've already instantited the Task object and have checked that
       the calling host has nvml GPU

       Parameters
       ==========
       skip: an optional list of (comma separated) fields to skip. Can be in
             net_io_counters,net_connections,net_if_address,net_if_stats
    '''
    nvmlInit()

    results = {}

    # A comma separated list of parameters to not include
    skip = kwargs.get('skip', '')
    skip = skip.split(',')

    # Run through high level metrics
    funcs = {'nvml_driver_version': pynvml.nvmlSystemGetDriverVersion,
             'nvml_system_nvml_version': pynvml.nvmlSystemGetNVMLVersion,
             'nvml_deviceCount': pynvml.nvmlDeviceGetCount,
             'nvml_hic_version': pynvml.nvmlSystemGetHicVersion,
             'nvml_unit_count': pynvml.nvmlUnitGetCount}

    for name, func in funcs.items():
        try:
            results[name] = func()
        except:
            nvmlInit()

    # Look at individual devices
    funcs = {
        'nvml_device_board_id': pynvml.nvmlDeviceGetBoardId,
        'nvml_device_multi_gpu_board': pynvml.nvmlDeviceGetMultiGpuBoard,
        'nvml_device_brand': pynvml.nvmlDeviceGetBrand,
        'nvml_device_serial': pynvml.nvmlDeviceGetSerial,
        'nvml_device_set_cpu_affinite': pynvml.nvmlDeviceSetCpuAffinity,
        'nvml_device_minor_number': pynvml.nvmlDeviceGetMinorNumber,
        'nvml_device_uuid': pynvml.nvmlDeviceGetUUID,
        'nvml_device_inforom_version': pynvml.nvmlDeviceGetInforomImageVersion,               
        'nvml_device_inforam_checksum': pynvml.nvmlDeviceGetInforomConfigurationChecksum,
        'nvml_device_display_mode': pynvml.nvmlDeviceGetDisplayMode,
        'nvml_device_display_active': pynvml.nvmlDeviceGetDisplayActive,
        'nvml_device_persistence_mode': pynvml.nvmlDeviceGetPersistenceMode,
        'nvml_device_supported_memory_clocks': pynvml.nvmlDeviceGetSupportedMemoryClocks,
        'nvml_device_fan_speed': pynvml.nvmlDeviceGetFanSpeed,
        'nvml_device_performance_state': pynvml.nvmlDeviceGetPerformanceState,
        'nvml_device_management_mode': pynvml.nvmlDeviceGetPowerManagementMode,
        'nvml_device_power_managerment_mode': pynvml.nvmlDeviceGetPowerManagementMode,
        'nvml_device_power_management_limit': pynvml.nvmlDeviceGetPowerManagementLimit,
        'nvml_device_power_management_limit_constraints': pynvml.nvmlDeviceGetPowerManagementLimitConstraints,
        'nvml_device_power_management_default_limit': pynvml.nvmlDeviceGetPowerManagementDefaultLimit,
        'nvml_device_enforced_power_limit': pynvml.nvmlDeviceGetEnforcedPowerLimit,
        'nvml_device_power_usage': pynvml.nvmlDeviceGetPowerUsage,
        'nvml_device_gpu_operation_mode': pynvml.nvmlDeviceGetGpuOperationMode,
        'nvml_device_current_operation_mode': pynvml.nvmlDeviceGetCurrentGpuOperationMode,
        'nvml_device_pending_gpu_operation_mode': pynvml.nvmlDeviceGetPendingGpuOperationMode,
        'nvml_device_memory_info': pynvml.nvmlDeviceGetMemoryInfo,
        'nvml_device_bar1_memory_info': pynvml.nvmlDeviceGetBAR1MemoryInfo,
        'nvml_device_compute_mode': pynvml.nvmlDeviceGetComputeMode,
        'nvml_device_ecc_mode': pynvml.nvmlDeviceGetEccMode,
        'nvml_device_current_ecc_mode': pynvml.nvmlDeviceGetCurrentEccMode,
        'nvml_device_pending_ecc_mode': pynvml.nvmlDeviceGetPendingEccMode,
        'nvml_device_utilization_rates': pynvml.nvmlDeviceGetUtilizationRates,
        'nvml_device_encoder_utilization': pynvml.nvmlDeviceGetEncoderUtilization,
        'nvml_device_decoder_utilization': pynvml.nvmlDeviceGetDecoderUtilization,
        'nvml_device_pci_replay_counter': pynvml.nvmlDeviceGetPcieReplayCounter,
        'nvml_device_driver_model': pynvml.nvmlDeviceGetDriverModel,
        'nvml_device_current_driver_model': pynvml.nvmlDeviceGetCurrentDriverModel,
        'nvml_device_pending_driver_model': pynvml.nvmlDeviceGetPendingDriverModel,
        'nvml_device_vbios_version': pynvml.nvmlDeviceGetVbiosVersion,
        'nvml_device_compute_running_processes': pynvml.nvmlDeviceGetComputeRunningProcesses,
        'nvml_device_grapics_running_processes': pynvml.nvmlDeviceGetGraphicsRunningProcesses,
        'nvml_device_auto_boosted_clocks_enabled': pynvml.nvmlDeviceGetAutoBoostedClocksEnabled,
        'nvml_device_supported_event_types': pynvml.nvmlDeviceGetSupportedEventTypes,
        'nvml_device_current_pcie_link_generation': pynvml.nvmlDeviceGetCurrPcieLinkGeneration,
        'nvml_device_max_pcie_link_generation': pynvml.nvmlDeviceGetMaxPcieLinkGeneration,
        'nvml_device_curr_pcie_link_width': pynvml.nvmlDeviceGetCurrPcieLinkWidth,
        'nvml_device_max_pcie_link_width': pynvml.nvmlDeviceGetMaxPcieLinkWidth,
        'nvml_device_supported_clocks_throttle_reasons': pynvml.nvmlDeviceGetSupportedClocksThrottleReasons,
        'nvml_device_current_clocks_throttle_reasons': pynvml.nvmlDeviceGetCurrentClocksThrottleReasons,
        'nvml_device_index': pynvml.nvmlDeviceGetIndex,
        'nvml_device_accounting_mode': pynvml.nvmlDeviceGetAccountingMode,
        'nvml_device_accounting_pids': pynvml.nvmlDeviceGetAccountingPids,
        'nvml_device_accounting_buffer_size': pynvml.nvmlDeviceGetAccountingBufferSize
    }

    # Functions that need additional args
    # nvmlDeviceGetCpuAffinity(handle, cpuSetSize):
    # nvmlDeviceGetInforomVersion(handle, infoRomObject)
    # nvmlDeviceGetClockInfo(handle, type)
    # nvmlDeviceGetMaxClockInfo(handle, type)
    # nvmlDeviceGetApplicationsClock(handle, type)
    # nvmlDeviceGetDefaultApplicationsClock(handle, type)
    # nvmlDeviceGetSupportedGraphicsClocks(handle, memoryClockMHz)
    # nvmlDeviceGetTemperature(handle, sensor)
    # nvmlDeviceGetTemperatureThreshold(handle, threshold)
    # nvmlDeviceGetTotalEccErrors(handle, errorType, counterType)
    # nvmlDeviceGetMemoryErrorCounter(handle, errorType, counterType, locationType)
    # nvmlDeviceRegisterEvents(handle, eventTypes, eventSet):
    # nvmlEventSetWait(eventSet, timeoutms):
    # nvmlDeviceOnSameBoard(handle1, handle2):
    # nvmlDeviceGetAccountingStats(handle, pid):

    # Setting functions that return None
    # nvmlDeviceSetCpuAffinity(handle)
    # nvmlDeviceClearCpuAffinity(handle)
    # nvmlDeviceValidateInforom(handle)        
    # nvmlUnitSetLedState(unit, color):
    # nvmlDeviceSetPersistenceMode(handle, mode):
    # nvmlDeviceSetComputeMode(handle, mode):    
    # nvmlDeviceSetEccMode(handle, mode):
    # nvmlDeviceClearEccErrorCounts(handle, counterType):
    # nvmlDeviceSetDriverModel(handle, model):    
    # nvmlDeviceSetAutoBoostedClocksEnabled(handle, enabled): 
    # nvmlDeviceSetDefaultAutoBoostedClocksEnabled(handle, enabled, flags): 
    # nvmlDeviceSetApplicationsClocks(handle, maxMemClockMHz, maxGraphicsClockMHz):    
    # nvmlDeviceResetApplicationsClocks(handle):
    # nvmlDeviceSetPowerManagementLimit(handle, limit):
    # nvmlDeviceSetGpuOperationMode(handle, mode):
    # nvmlDeviceSetAccountingMode(handle, mode):
    # nvmlDeviceClearAccountingPids(handle):

    device_count = results['nvml_deviceCount']

    devices = {}
    for i in range(device_count):
        handle = pynvml.nvmlDeviceGetHandleByIndex(i)
        name = pynvml.nvmlDeviceGetName(handle)
        devices[name] = {}
        for key, func in funcs.items():
            try:

                result = func(handle)

                # Special parsing of the result depending on the type
                if isinstance(result, map):
                    result = list(result)

                if key == 'nvml_device_bar1_memory_info':
                    result = {"bar1Free": result.bar1Free,
                              "bar1Total": result.bar1Total,
                              "bar1Used": result.bar1Used}

                if key == 'nvml_device_utilization_rates':
                    result = {"gpu": result.gpu,
                              "memory": result.memory}

                elif key == 'nvml_device_memory_info':
                    result = {"free": result.free,
                              "total": result.total,
                              "used": result.used}

                devices[name][key] = result
            except:
                nvmlInit()
   
    nvmlShutdown()

    results['devices'] = devices
    return _filter_result(results, skip)


def _filter_result(results, skip):
    '''a helper function to filter a dictionary based on a list of keys to 
       skip. We also add variables from the environment.
    
       Parameters
       ==========
       results: a dictionary of results
       skip: a list of keys to remove/filter from the result.
    '''

    # Add any environment variables prefixed wit WATCHMEENV_
    environ = get_watchme_env()
    results.update(environ)

    for key in skip:
        if key in results:
            del results[key]

    return results



'''

Copyright (C) 2019 Vanessa Sochat.

This Source Code Form is subject to the terms of the
Mozilla Public License, v. 2.0. If a copy of the MPL was not distributed
with this file, You can obtain one at http://mozilla.org/MPL/2.0/.

'''

from watchme.utils import ( run_command, mkdir_p )
from watchme.defaults import ( 
    WATCHME_WATCHER, 
    WATCHME_BASE_DIR
)
from watchme.logger import bot
from watchme.config import generate_watcher_config
import os


def create_watcher(name=None, watcher_type=None, base=None):
    '''create a watcher, meaning a folder with a configuration and
       initialized git repo.

       Parameters
       ==========
       name: the watcher to create, uses default or WATCHME_WATCHER
       watcher_type: the type of watcher to create. defaults to 
                     WATCHER_DEFAULT_TYPE
        base: The watcher base to use (defaults to $HOME/.watchme)
    '''
    if name is None:
        name = WATCHME_WATCHER

    if base is None:
        base = WATCHME_BASE_DIR

    # Create the repository folder
    repo = os.path.join(base, name)

    if not os.path.exists(repo):

        bot.info('Adding watcher %s...' % repo)
        mkdir_p(repo)

        # Ensure no gpg signing happens
        run_command("git --git-dir=%s/.git init" % repo)
        run_command("git --git-dir=%s/.git config commit.gpgsign false" % repo)
    
        # Add the watcher configuration file
        generate_watcher_config(repo, watcher_type)
        run_command("git -C %s add watchme.cfg" % repo)  
        return repo

    else:
        bot.info('%s already exists: %s' % (name, repo))


def create_watcher_base(name=None, base=None):
    '''create a watch base and default repo, if it doesn't already exist.

       Parameters
       ==========
       name: the watcher to create, uses default or WATCHME_WATCHER
       base: the watcher base, defaults to WATCHME_BASE_DIR
    '''
    if base is None:
        base = WATCHME_BASE_DIR

    if name is None:
        name = WATCHME_WATCHER

    if not os.path.exists(base):
        bot.info('Creating %s...' % base)
        mkdir_p(base)



'''

Copyright (C) 2019 Vanessa Sochat.

This Source Code Form is subject to the terms of the
Mozilla Public License, v. 2.0. If a copy of the MPL was not distributed
with this file, You can obtain one at http://mozilla.org/MPL/2.0/.

'''


from subprocess import (
    Popen,
    PIPE,
    STDOUT
)
import os
import re
import shlex

# Terminal Commands

def which(software, strip_newline=True):
    '''get_install will return the path to where an executable is installed.
    '''
    if software is None:
        software = "watchme"
    cmd = 'which %s' % software
    try:
        result = run_command(cmd)
        if strip_newline is True:
            result['message'] = result['message'].strip('\n')
        if "message" in result:
            return result['message']
        return result

    except: # FileNotFoundError
        return None


def get_installdir():
    '''get_installdir returns the installation directory of the application
    '''
    return os.path.abspath(os.path.dirname(os.path.dirname(__file__)))


def run_command(cmd, sudo=False):
    '''run_command uses subprocess to send a command to the terminal.

    Parameters
    ==========
    cmd: the command to send, should be a list for subprocess
    error_message: the error message to give to user if fails,
    if none specified, will alert that command failed.

    '''

    cmd = shlex.split(cmd)

    if sudo is True:
        cmd = ['sudo'] + cmd

    try:
        output = Popen(cmd, stderr=STDOUT, stdout=PIPE)

    except FileNotFoundError:
        cmd.pop(0)
        output = Popen(cmd,stderr=STDOUT,stdout=PIPE)

    t = output.communicate()[0],output.returncode
    output = {'message':t[0],
              'return_code':t[1]}

    if isinstance(output['message'], bytes):
        output['message'] = output['message'].decode('utf-8')

    return output


# Environment

def convert2boolean(arg):
    '''
    convert2boolean is used for environmental variables
    that must be returned as boolean
    '''
    if not isinstance(arg, bool):
        return arg.lower() in ("yes", "true", "t", "1", "y")
    return arg


def get_watchme_env(prefix="WATCHMEENV_"):
    '''get any environment variables that start with WATCMEENV_, return the
       dictionary to the user.
    '''
    environ = {}

    # First extract variables from the environment
    for key, value in os.environ.items():

        # Variables that are specified, or start with WATCHMEENV included
        if key.startswith(prefix):

            # Replace the WATCHMEENV_ if present
            key = re.sub("^%s" % prefix, "", key)

            # Don't include empty strings
            if value not in ["", None]:

                environ[key] = value

    return environ



'''

Copyright (C) 2019 Vanessa Sochat.

This Source Code Form is subject to the terms of the
Mozilla Public License, v. 2.0. If a copy of the MPL was not distributed
with this file, You can obtain one at http://mozilla.org/MPL/2.0/.

'''

from watchme.command import (
    get_watchers,
    list_task,
    list_watcher,
    list_watcher_types
)
from watchme.logger import bot

def main(args, extra):
    '''list installed watchers
    '''
    if args.watchers is True:
        list_watcher_types()

    # Otherwise, we are listing installed watchers and tasks
    else:

        # If no watchers provided, list the watchers
        if extra is None:
            get_watchers(args.base)

        # One argument is the name of a watcher
        elif len(extra) == 1:
            list_watcher(extra[0], args.base)

        # Two arguments must be a watcher and task
        elif len(extra) == 2:
            list_task(extra[0], extra[1], args.base)

        else:
            bot.exit('Please provide none or all of <watcher> <task> to list.')



'''

Copyright (C) 2019 Vanessa Sochat.

This Source Code Form is subject to the terms of the
Mozilla Public License, v. 2.0. If a copy of the MPL was not distributed
with this file, You can obtain one at http://mozilla.org/MPL/2.0/.

'''

from watchme import get_watcher
from watchme.logger import bot

def main(args, extra):
    '''activate one or more watchers
    '''    
    # Required - will print help if not provided
    name = args.watcher[0]
    watcher = get_watcher(name, base=args.base, create=False)

    # If delete is true, remove entire watcher, only if not protected or frozen
    if args.delete:
        watcher.delete()    

    else:

        # Exit if the user doesn't provide any tasks to remove
        if extra is None:
            bot.exit('Provide tasks to remove, or --delete for entire watcher.')
    
        for task in extra:
     
            # Remove the task, if it exists
            watcher.remove_task(task)



'''

Copyright (C) 2019 Vanessa Sochat.

This Source Code Form is subject to the terms of the
Mozilla Public License, v. 2.0. If a copy of the MPL was not distributed
with this file, You can obtain one at http://mozilla.org/MPL/2.0/.

'''

from subprocess import (
    check_output, 
    CalledProcessError
)
from watchme.utils import get_watchme_env
from watchme.logger import bot
import psutil

def monitor_pid_task(**kwargs):
    '''monitor a specific process. This function can be used as a task, or
       is (most likely used) for the psutils.decorators. A pid parameter
       is required.

       Parameters
       ==========
       skip: an optional list of (comma separated) fields to skip. Can be in
             net_io_counters,net_connections,net_if_address,net_if_stats
       pid: the process id or name (required)
    '''
    pid = kwargs.get('pid', None)

    bot.debug(kwargs)

    # Helper function to get list from one,two,three
    def get_list(name):
        csv_list = kwargs.get(name, '')
        return [x for x in csv_list.split(',') if x]

    # A comma separated list of parameters to skip
    skip = get_list('skip')
    include = get_list('include')
    only = get_list('only')

    # Only continue given that argument is provided
    if pid is None:   
        bot.warning("A 'pid' parameter is required to use the monitor_pid_task function.")
        return pid

    # The user is allowed to provide a process name, or a number
    try:
        pid = int(pid)
    except ValueError:
        pid = _get_pid(pid)

    # But if it's stil no good (None) we exit.
    if pid is None:
        bot.warning("'pid' must be a running process or process name.")
        return pid

    ps = psutil.Process(pid)
    bot.debug(ps)

    results = {}

    for key, val in ps.as_dict().items():
 
        # If val is None, don't include
        if val is None:
            bot.debug('skipping %s, None' % key)
            continue

        if key == "connections":
            connections = []
            for net in val:
                entry = {'fd': net.fd,
                         'family': str(net.family),
                         'type': str(net.type),
                         'laddr_ip': net.laddr.ip,
                         'laddr_port': net.laddr.port,
                         'raddr': net.raddr,
                         'status': net.status}
                connections.append(entry)
            val = connections       

        # First priority goes to a custom set
        if len(only) > 0:
            if key in only:
                results[key] = val
            else:
                bot.debug('skipping %s' % key)
            continue 

        # The user requested to skip
        elif key in skip or key in ['threads']:
            continue

        # Keep count of openfiles
        elif key in ["open_files"]:
            results[key] = len(val)

        # Don't risk exposing sensitive information
        elif key == "environ":
            if key in include:
                results[key] = val

        # Skip over ones that are too detailed
        elif key in ['memory_maps']:
            continue

        # I assume this included all that are in pmem too
        elif isinstance(val, psutil._pslinux.pfullmem):
            results[key] = {"rss": val.rss,
                            "vms": val.vms,
                            "shared": val.shared,
                            "text": val.text,
                            "lib": val.lib,
                            "data": val.data,
                            "dirty": val.dirty,
                            "uss": val.uss,
                            "pss": val.pss,
                            "swap": val.swap}

        elif isinstance(val, psutil._common.pgids) or isinstance(val, psutil._common.puids):
            results[key] = {"real": val.real,
                            "effetive":val.effective,
                            "saved": val.saved}

        elif isinstance(val, psutil._common.pcputimes):
            results[key] = {"user": val.user,
                            "system": val.system,
                            "children_user": val.children_user,
                            "children_system": val.children_system}

        elif isinstance(val, psutil._common.pctxsw):
            results[key] = {"voluntary": val.voluntary,
                            "involuntary":val.involuntary}

        elif isinstance(val, psutil._common.pionice):
            results[key] = {"value": val.value}

            # Older Python version (2) doesn't have attribute
            if hasattr(val.ioclass, 'name'):
                results[key]["ioclass"] = val.ioclass.name

        # pfullmem (first above) should cover this
        elif isinstance(val, psutil._pslinux.pmem):
            continue

        elif isinstance(val, psutil._pslinux.pio):
            results[key] = {"read_count": val.read_count,
                            "write_count": val.write_count,
                            "read_bytes": val.read_bytes,
                            "write_bytes": val.write_bytes,
                            "read_chars": val.read_chars,
                            "write_chars": val.write_chars}
        else:
            results[key] = val

    # Add any environment variables prefixed wit WATCHMEENV_
    environ = get_watchme_env()
    results.update(environ)

    return results

def cpu_task(**kwargs):
    '''Get variables about the cpu of the host. No parameters are required.

       Parameters
       ==========
       skip: an optional list of (comma separated) fields to skip. Can be in
             net_io_counters,net_connections,net_if_address,net_if_stats
    '''
    result = {}

    # A comma separated list of parameters to not include
    skip = kwargs.get('skip', '')
    skip = skip.split(',')

    result['cpu_freq'] = dict(psutil.cpu_freq()._asdict())
    result['cpu_percent'] = psutil.cpu_percent()
    result['cpu_count'] = psutil.cpu_count()
    result['cpu_stats'] = dict(psutil.cpu_stats()._asdict())
    result['cpu_times'] = dict(psutil.cpu_times()._asdict())
    result['cpu_times_percent'] = dict(psutil.cpu_times_percent()._asdict())
    return _filter_result(result, skip)


def python_task(**kwargs):
    '''Get modules, version, etc. about currently running python

       Parameters
       ==========
       skip: an optional list of (comma separated) fields to skip. Can be in
             net_io_counters,net_connections,net_if_address,net_if_stats
    '''
    # A comma separated list of parameters to not include
    skip = kwargs.get('skip', '')
    skip = skip.split(',')

    result = {'modules': list(psutil.os.sys.modules),
              'version': psutil.os.sys.version,
              'path': psutil.os.sys.path,
              'prefix': psutil.os.sys.prefix,
              'executable': psutil.os.sys.executable,
              'copyright': psutil.os.sys.copyright }

    return _filter_result(result, skip)


def system_task(**kwargs):
    '''Get basic system info, unlikely to change.

       Parameters
       ==========
       skip: an optional list of (comma separated) fields to skip. Can be in
             net_io_counters,net_connections,net_if_address,net_if_stats
    '''
    # A comma separated list of parameters to not include
    skip = kwargs.get('skip', '')
    skip = skip.split(',')

    result = {}
    result['platform'] = psutil.os.sys.platform
    result['api_version'] = psutil.os.sys.api_version
    result['maxsize'] = psutil.os.sys.maxsize

    # Python 2 doesn't have these fields
    if hasattr(psutil.os.sys, 'int_info'):
        result['bits_per_digit'] = psutil.os.sys.int_info.bits_per_digit
        result['sizeof_digit'] = psutil.os.sys.int_info.sizeof_digit
 
    vinfo = psutil.os.sys.version_info
    result['version_info'] = {'major': vinfo.major,
                              'minor': vinfo.minor,
                              'micro': vinfo.micro,
                              'releaselevel': vinfo.releaselevel,
                              'serial': vinfo.serial}

    result['sep'] = psutil.os.sep
    uname = psutil.os.uname()
    
    # Python 2 returns a tuple 
    if not hasattr(uname, 'sysname'):
        result['uname'] = uname
    else:
        result['uname'] = {'sysname': uname.sysname,
                           'nodename': uname.nodename,
                           'version': uname.version,
                           'release': uname.release,
                           'machine': uname.machine}

    return _filter_result(result, skip)


def memory_task(**kwargs):
    '''Get values for memory of the host. No parameters are required.
    '''

    result = {}

    # gives an object with many fields
    result['virtual_memory'] = dict(psutil.virtual_memory()._asdict())

    # Add any environment variables prefixed wit WATCHMEENV_
    environ = get_watchme_env()
    result.update(environ)

    return result


def users_task(**kwargs):
    '''Get values for current users
    '''

    result = {'users': []}
    for user in psutil.users():
        result['users'].append(dict(user._asdict()))

    # Add any environment variables prefixed wit WATCHMEENV_
    environ = get_watchme_env()
    result.update(environ)

    return result


def sensors_task(**kwargs):
    '''Get values from system sensors like fans, temperature, battery
    '''

    # A comma separated list of parameters to not include
    skip = kwargs.get('skip', '')
    skip = skip.split(',')

    result = {}
    result['sensors_temperatures'] = {}
    result['sensors_fans'] = {}

    # battery
    battery = psutil.sensors_battery()
    if battery is not None:
        result['sensors_battery'] = {'percent': battery.percent,
                                     'secsleft': str(battery.secsleft),
                                     'power_plugged': battery.power_plugged}

    # fans
    for name, entry in psutil.sensors_fans().items():
        entries = []
        for e in entry:
            item = {'label': e.label,
                    'current': e.current}
            entries.append(item)
        result['sensors_fans'][name] = entries

    # temperature
    for name, entry in psutil.sensors_temperatures().items():
        entries = []
        for e in entry:
            temp = {'label': e.label,
                    'current': e.current,
                    'high': e.high,
                    'critical': e.critical}
            entries.append(temp)
        result['sensors_temperatures'][name] = entries

    return _filter_result(result, skip)



def net_task(**kwargs):
    '''Get values for networking on the host

       Parameters
       ==========
       skip: an optional list of (comma separated) fields to skip. Can be in
             net_io_counters,net_connections,net_if_address,net_if_stats
    '''
    result = {}

    # A comma separated list of parameters to not include
    skip = kwargs.get('skip', '')
    skip = skip.split(',')

    result['net_connections'] = []
    result['net_if_address'] = {}
    result['net_if_stats'] = {}

    for net in psutil.net_connections():
        net_result = {'fd': net.fd,
                      'family': str(net.family),
                      'type': str(net.type),
                      'laddr_ip': net.laddr.ip,
                      'laddr_port': net.laddr.port,
                      'raddr': net.raddr,
                      'status': net.status,
                      'pid': net.pid}
        result['net_connections'].append(net_result)

    for name, entry in psutil.net_if_addrs().items():
        entries = []
        for e in entry:
            item = {'family': str(e.family),
                    'address': e.address,
                    'netmask': e.netmask,
                    'broadcast': e.broadcast,
                    'ptp': e.ptp}
            entries.append(item)
        result['net_if_address'][name] = entries

    for name, entry in psutil.net_if_stats().items():
        item = {'isup': entry.isup,
                'duplex': str(entry.duplex),
                'speed': entry.speed,
                'mtu': entry.mtu}
        result['net_if_stats'][name] = item

    netio = psutil.net_io_counters()
    result['net_io_counters'] = {'bytes_sent': netio.bytes_sent,
                                 'bytes_recv': netio.bytes_recv,
                                 'packets_sent': netio.packets_sent,
                                 'packets_recv': netio.packets_recv,
                                 'errin': netio.errin,
                                 'errout': netio.errout,
                                 'dropin': netio.dropin,
                                 'dropout': netio.dropout}

    # Remove those not wanted
    return _filter_result(result, skip)


def disk_task(**kwargs):
    '''Get values for disks on the host

       Parameters
       ==========
       disk_usage: an optional path to get disk usage for.
       skip: an optional list of (comma separated) fields to skip. Can be in
             net_io_counters,net_connections,net_if_address,net_if_stats

    '''
    # A comma separated list of parameters to not include
    skip = kwargs.get('skip', '')
    skip = skip.split(',')
    path = kwargs.get('disk_usage', '/')

    result = {}
    result['disk_to_counters'] = dict(psutil.disk_io_counters()._asdict())
    result['disk_partitions'] = []
    for disk in psutil.disk_partitions():
        disk_result = {'device': disk.device,
                       'mountpoint': disk.mountpoint,
                       'fstype': disk.fstype,
                       'opts': disk.opts}
        result['disk_partitions'].append(disk_result)
    result['disk_usage'] = dict(psutil.disk_usage(path)._asdict())

    # Remove those not wanted
    return _filter_result(result, skip)


def _filter_result(result, skip):
    '''a helper function to filter a dictionary based on a list of keys to 
       skip. We also add variables from the environment.
    
       Parameters
       ==========
       result: a dictionary of results
       skip: a list of keys to remove/filter from the result.
    '''

    # Add any environment variables prefixed wit WATCHMEENV_
    environ = get_watchme_env()
    result.update(environ)

    for key in skip:
        if key in result:
            del result[key]

    return result


def _get_pid(name):
    '''used to get the pid of a process, by name
       
       Parameters
       ==========
       name: the name of the process to get.
    '''
    try:
        pid = check_output(["pidof", name])
        pid = pid.decode('utf-8').strip('\n').split(' ') 
        if len(pid) > 1:
            bot.warning("More than one pid found for %s, using first." % name)
        pid = int(pid[0])

    except CalledProcessError:
        bot.error("%s does not exist." % name)
        pid = None
    return pid



'''

Copyright (C) 2019 Vanessa Sochat.

This Source Code Form is subject to the terms of the
Mozilla Public License, v. 2.0. If a copy of the MPL was not distributed
with this file, You can obtain one at http://mozilla.org/MPL/2.0/.

'''

from functools import wraps
from watchme.logger import bot
from watchme.tasks.decorators import ProcessRunner
from watchme import get_watcher


def monitor_gpu(*args, **kwargs):
    '''a decorator to monitor a function every 3 (or user specified) seconds. 
       We include one or more task names that include data we want to extract.
       we get the pid of the running function, and then use the
       gpu_task from gpu to watch it. The functools "wraps"
       ensures that the (fargs, fkwargs) are passed from the calling function
       despite the wrapper. The following parameters can be provided to
       "monitor resources"

       Parameters
       ==========
       watcher: the watcher instance to use, used to save data to a "task"
                folder that starts with "decorator-<name<"
       seconds: how often to collect data during the run.
       only: ignore skip and include, only include this custom subset
       skip: Fields in the result to skip (list).
       include: Fields in the result to include back in (list).
       create: whether to create the watcher on the fly (default False, must
               exist)
       name: the suffix of the decorator-gpu-<name> folder. If not provided,
             defaults to the function name
    '''
    def inner(func):

        @wraps(func)
        def wrapper(*fargs, **fkwargs):

            # Typically the task folder is the index, so we will create
            # indices that start with decorator-<task>
            result = None

            # The watcher is required, first keyword argument
            if not args:
                bot.error("A watcher name is required for the gpu decorator.")
                return result

            # Get a watcher to save results to
            watcher = get_watcher(args[0], create=kwargs.get('create', False))

            # Start the function
            runner = ProcessRunner(seconds=kwargs.get('seconds', 3),
                                   skip=kwargs.get('skip', []),
                                   include=kwargs.get('include', []),
                                   only=kwargs.get('only', []))

            runner.run(func, *fargs, **fkwargs)
            result = runner.wait("gpu_task")

            # Save results (finishing runs) - key is folder created
            name = kwargs.get('name', func.__name__)
            key = 'decorator-gpu-%s' % name
            results = {key: runner.timepoints}
            watcher.finish_runs(results)
            
            # Return function result to the user
            return result
        return wrapper
    return inner



'''

Copyright (C) 2019 Vanessa Sochat.

This Source Code Form is subject to the terms of the
Mozilla Public License, v. 2.0. If a copy of the MPL was not distributed
with this file, You can obtain one at http://mozilla.org/MPL/2.0/.

'''

from watchme.command import create_watcher

def main(args, extra):
    '''create means creating one or more watchers.
    '''    

    watchers = args.watchers
    if len(watchers) == 0:
        watchers = ['watcher']

    for watcher in watchers:
        create_watcher(watcher)



'''

Copyright (C) 2019 Vanessa Sochat.

This Source Code Form is subject to the terms of the
Mozilla Public License, v. 2.0. If a copy of the MPL was not distributed
with this file, You can obtain one at http://mozilla.org/MPL/2.0/.


'''

from .pynvml import nvmlInit
from watchme.tasks import TaskBase
from watchme.logger import bot
from watchme.utils import (
    get_user, 
    get_host
)
import os
import sys

class Task(TaskBase):

    required_params = []

    def __init__(self, name, params=None, **kwargs): 

        if params is None:
            params = {}

        self.type = 'gpu'
        self.assert_gpu()

        # If the user doesn't provide a file name, name based on task
        if "_save" in kwargs:
            if "file_name" not in params:
                params['file_name'] = "%s_%s.json" % (get_host().lower(), 
                                                      get_user().lower())

        # Handles setting the name, setting params, and validate
        super(Task, self).__init__(name, params, **kwargs)

    def assert_gpu(self):
        '''has_gpu is run from the getgo to see if there are any libraries
           for the client to read from. If not, we alert the user and exit.
        '''
        try:
            nvmlInit()
        except:
            bot.exit("NVML Shared Library Not Found.")

    def export_func(self):
        '''this function should return the correct task (from the tasks.py
           in the same folder) based on some logic of the params that are given
           by the user (self.params). If there is only one kind of function for
           the task, it's fairly easy to import and return it here. This
           function should take no arguments, but instead use the self.params
           already provided in the client.
        '''
        name = self.params.get('func', 'gpu_task')

        if name == 'gpu_task':
            from .tasks import gpu_task as func
        else:
            func = None

        return func



'''

Copyright (C) 2019 Vanessa Sochat.

This Source Code Form is subject to the terms of the
Mozilla Public License, v. 2.0. If a copy of the MPL was not distributed
with this file, You can obtain one at http://mozilla.org/MPL/2.0/.

'''

from watchme import get_watcher


def main(args, extra):
    '''activate one or more watchers
    '''    
    # Required - will print help if not provided
    name = args.watcher[0]
    watcher = get_watcher(name, base=args.base, create=False)

    # If user wants to see create command, must have list of tasks
    if args.create_command is True and extra is None:
        extra = [x.name for x in watcher.get_tasks()]

    # Show the create command, or inspect a task
    watcher.inspect(extra, create_command=args.create_command)



'''

Copyright (C) 2019 Vanessa Sochat.

This Source Code Form is subject to the terms of the
Mozilla Public License, v. 2.0. If a copy of the MPL was not distributed
with this file, You can obtain one at http://mozilla.org/MPL/2.0/.

'''

from watchme.utils import (
    get_tmpdir,
    get_watchme_env,
    write_file
)
import os
import shutil


def from_env_task(**kwargs):
    '''Get some set of variables from the environment. We look for those
       defined in kwargs, but also any in the environment that start with
       the prefix WATCHMENEV_. They are saved to files that are equivalently
       named, and the idea is that we can track a changing (single) result,
       meaning that the timestamp is meaningful, or we can track
       values for many different results, so the timestamps just serve to
       record when each was recorded.

       Parameters
       ==========
       *: any number of parameters from the task configuration that will
          be looked for in the environment.
    '''
    results = []
 
    # Create a temporary directory for results
    tmpdir = get_tmpdir()

    # First extract variables from the environment
    environ = get_watchme_env()
    for key, value in environ.items():
 
        # Write the result to file (don't include extension)
        filename = os.path.join(tmpdir, key)
        write_file(filename, value)
        results.append(filename)

    # If no results, return None
    if len(results) == 0:
        shutil.rmtree(tmpdir)

    return results



'''

Copyright (C) 2019 Vanessa Sochat.

This Source Code Form is subject to the terms of the
Mozilla Public License, v. 2.0. If a copy of the MPL was not distributed
with this file, You can obtain one at http://mozilla.org/MPL/2.0/.

'''

from multiprocessing import (
    Process, 
    Queue
)
from watchme.logger import bot
from time import sleep
import shlex
import subprocess
import os

class DecoratorBase(object):

    def __init__(self, seconds=3, skip=None, include=None, only=None):
 
        # Replace Nones with lists
        only = none_to_list(only)
        include = none_to_list(include)
        skip = none_to_list(skip)

        self.seconds = seconds
        self.timepoints = []

        # Export seconds to the environment
        os.environ["WATCHMEENV_SECONDS"] = str(self.seconds)

        # Ensure we have csv lists
        self.only = self._parse_custom(only)
        self.skip = self._parse_custom(skip)
        self.include = self._parse_custom(include)

    def _parse_custom(self, listy):
        '''parse an actual list (['one','two','three']) into 
           a csv list. If we don't have a list, ignore and assume already
           parsed that way.
 
           Parameters
           ==========
           listy: the actual list
        '''
        if listy is None:
            listy = []
        if isinstance(listy, list):
            listy = ','.join(listy)
        return listy

    def run(self, *args, **kwargs):
        '''run should be implemented by the subclass to run the function or
           process being monitored
        '''
        bot.exit('run function must be implemented by subclass.')

    def wait(self, *args, **kwargs):
        '''wait should be run after run to monitor the process being run.'''
        bot.exit('wait function must be implemented by subclass.')


class TerminalRunner(DecoratorBase):

    def __init__(self, cmd, **kwargs):
        self.cmd = shlex.split(cmd)
        self.process = None

        # Handles setting the skip, include, and only parameters
        super(TerminalRunner, self).__init__(**kwargs)

    def run(self):
        '''run the user provided function
        '''
        try:
            self.process = subprocess.Popen(self.cmd, stderr=subprocess.STDOUT, stdout=subprocess.PIPE)
        except FileNotFoundError:
            self.cmd.pop(0)
            self.process = subprocess.Popen(self.cmd, stderr=subprocess.STDOUT, stdout=subprocess.PIPE)

    def wait(self, task_name):
        '''wait should monitor the running task. run should be called first.

           Parameters
           ==========
           task_name: should correspond with the task func (e.g., 
                      monitor_pid_task for psutils, or
                      monitor_gpu for gpu tasks.
        '''

        # Parameters for the pid, and to skip sections of results
        params = {"skip": self.skip,
                  "pid": self.process.pid,
                  "include": self.include,
                  "only": self.only,
                  'func': task_name}

        # This particular decorator doesn't take input params
        task = get_task(task_name, params)

        # Export parameters and functions            
        function = task.export_func()
        params = task.export_params()

        # collect resources, then sleep
        while self.process.poll() is None:

            # Function returns dictionary, we append to list of timepoints
            self.timepoints.append(function(**params))
            sleep(self.seconds)

        # Get the timepoints
        return self.timepoints


class ProcessRunner(DecoratorBase):

    def __init__(self, **kwargs):
        self.process = None
        self.queue = Queue()

        # Handles setting the seconds, skip, include, and only parameters
        super(ProcessRunner, self).__init__(**kwargs)

    @staticmethod
    def _wrapper(func, queue, args, kwargs):
        ret = func(*args, **kwargs)
        queue.put(ret)

    def run(self, func, *args, **kwargs):
        '''run the user provided function. When we run, we save the function
           run to self.func, where we can derive the name in self.wait.
        '''
        self.func = func
        args2 = [func, self.queue, args, kwargs]
        p = Process(target=self._wrapper, args=args2)
        self.process = p
        p.start()

    def wait(self, task_name):
        '''watch should monitor the running task. run should be called first.

           Parameters
           ==========
           task_name: should correspond with the task func (e.g., 
                      monitor_pid_task for psutils, or
                      monitor_gpu for gpu tasks.
        '''

        # Parameters for the pid, and to skip sections of results
        params = {"skip": self.skip,
                  "pid": self.process.pid,
                  "include": self.include,
                  "only": self.only,
                  'func': task_name}

        # This particular decorator doesn't take input params
        task = get_task(task_name, params)

        # Export parameters and functions            
        function = task.export_func()
        params = task.export_params()

        # collect resources, then sleep
        while self.process.is_alive():

            # Function returns dictionary, we append to list of timepoints
            self.timepoints.append(function(**params))
            sleep(self.seconds)

        # Get the result, and the timepoints
        result = self.queue.get()
        self.process.join()
        return result


def get_task(task_name, params):
    '''a helper function to return an instantiated task object, depending
       on the function name.

       Parameters
       ==========
       task_name: the name of the task to return
       params: parameters for the task
    '''
    if task_name == "monitor_pid_task":
        from watchme.watchers.psutils import Task
    elif task_name == "gpu_task":
        from watchme.watchers.gpu import Task
    else:
        bot.exit('%s is not a known task.' % task_name)
    return Task(task_name, params=params)


def none_to_list(value):
    '''a helper so that function args can be None, and updated to a list.
    '''
    if not value:
        value = []
    return value



'''

Copyright (C) 2019 Vanessa Sochat.

This Source Code Form is subject to the terms of the
Mozilla Public License, v. 2.0. If a copy of the MPL was not distributed
with this file, You can obtain one at http://mozilla.org/MPL/2.0/.

The watcher is actually a connection to crontab. This is what helps to schedule
the watched to check for changes at some frequency, and update the files.


'''

from watchme.logger import bot
from watchme.utils import get_userhome
import os
import sys


################################################################################
# environment / options
################################################################################


def getenv(variable_key, default=None, required=False, silent=True):
    ''' attempt to get an environment variable. If the variable
        is not found, None is returned.

        Parameters
        ==========
        variable_key: the variable name
        required: exit with error if not found
        silent: Do not print debugging information for variable
    '''
    variable = os.environ.get(variable_key, default)
    if variable is None and required:
        bot.error("Cannot find environment variable %s, exiting." % variable_key)
        sys.exit(1)

    if not silent and variable is not None:
        bot.verbose("%s found as %s" %(variable_key,variable))

    return variable



################################################################################
# Helpme

USERHOME = get_userhome()

WATCHME_WATCHER = getenv('WATCHME_WATCHER', 'watcher')
_config = os.path.join(USERHOME, ".watchme")
WATCHME_BASE_DIR = getenv('WATCHME_BASE_DIR', _config)
WATCHME_WORKERS = int(getenv('WATCHME_WORKERS', 9))

# The types of valid watchers (currently only urls). Corresponds with
# a folder under "main/watchers"
WATCHME_TASK_TYPES = ['urls', 'url', 'gpu', 'psutils', 'results']
WATCHME_DEFAULT_TYPE = "urls"

# Parameters not allowed for task clients, set by TaskBase
WATCHME_NOTALLOWED_PARAMS = ['type', 'active']



'''

Copyright (C) 2019 Vanessa Sochat.

This Source Code Form is subject to the terms of the
Mozilla Public License, v. 2.0. If a copy of the MPL was not distributed
with this file, You can obtain one at http://mozilla.org/MPL/2.0/.


'''

from watchme.tasks import TaskBase
from watchme.logger import bot
import os
import sys

class Task(TaskBase):

    required_params = ['url']

    def __init__(self, name, params=None, **kwargs): 

        if params is None:
            params = {}

        self.type = 'urls'

        # Handles setting the name, setting params, and validate
        super(Task, self).__init__(name, params, **kwargs)

    def _validate(self):
        '''additional validation function, called by validate() of 
           superclass. Here we assume all required self.params are included.
           If an parameter is found to be invalid, self.valid should be set
           to False
        '''
        # The url must begin with http
        if not self.params['url'].startswith('http'):
            bot.error('%s is not a valid url.' % self.params['url'])
            self.valid = False

    def export_func(self):
        '''this function should return the correct task (from the tasks.py
           in the same folder) based on some logic of the params that are given
           by the user (self.params). If there is only one kind of function for
           the task, it's fairly easy to import and return it here. This
           function should take no arguments, but instead use the self.params
           already provided in the client.
        '''
        name = self.params.get('func', 'get_task')

        if name == 'get_task':
            from .tasks import get_task as func
        elif name == 'download_task':
            from .tasks import download_task as func
        elif name == 'post_task':
            from .tasks import post_task as func
        elif name == "get_url_selection":
            from .tasks import get_url_selection as func
        else:
            func = None

        bot.debug('function name is %s' % name)
        return func



#!/usr/bin/python

# Copyright (C) 2019 Vanessa Sochat.

# This Source Code Form is subject to the terms of the
# Mozilla Public License, v. 2.0. If a copy of the MPL was not distributed
# with this file, You can obtain one at http://mozilla.org/MPL/2.0/.

from watchme import get_watcher
from watchme.command import create_watcher
import unittest
import tempfile
import shutil
import os


print("########################################################### test_client")


class TestClient(unittest.TestCase):

    def setUp(self):
        self.base = tempfile.mkdtemp()
        self.repo = create_watcher('pancakes', base=self.base)
        self.cli = get_watcher('pancakes', base=self.base)
        
    def tearDown(self):
        shutil.rmtree(self.base)

    def test_create(self):
        '''test creation function, and basic watcher config'''
        print("Testing watcher creation")
        self.assertTrue(os.path.exists(self.repo))        
        self.assertTrue(os.path.exists(os.path.join(self.repo, 'watchme.cfg')))  
        already_exists = create_watcher('pancakes', base=self.base)
        self.assertEqual(already_exists, None)
        self.assertEqual(self.cli.name, 'pancakes')
        self.assertEqual(self.cli.base, self.base)

    def test_activate(self):
        print("Testing watcher activation")
        self.assertEqual(self.cli.is_active(), False)
        self.cli.activate()
        self.assertEqual(self.cli.is_active(), True)
        self.cli.deactivate()
        self.assertEqual(self.cli.is_active(), False)

    def test_freeze(self):
        print("Testing watcher freezing")
        self.assertEqual(self.cli.is_frozen(), False)
        self.cli.freeze()
        self.assertEqual(self.cli.is_frozen(), True)
        self.cli.unfreeze()
        self.assertEqual(self.cli.is_frozen(), False)

        print("Testing watcher protect")
        self.assertEqual(self.cli.is_protected(), False)
        self.cli.protect("on")
        self.assertEqual(self.cli.is_protected(), True)
        self.cli.protect("off")
        self.assertEqual(self.cli.is_protected(), False)

    def test_watcher_task(self):
        '''test adding a task
        '''
        print("Testing watcher.add_task")
        self.cli.add_task('task-pancakes', 'urls', params=['url@https://www.google.com'])
        self.assertTrue(os.path.exists(os.path.join(self.repo, 'task-pancakes')))

        print("Testing task activation")
        self.assertEqual(self.cli.is_active('task-pancakes'), True)
        self.cli.deactivate('task-pancakes')
        self.assertEqual(self.cli.is_active('task-pancakes'), False)
        self.cli.activate('task-pancakes')
        self.assertEqual(self.cli.is_active('task-pancakes'), True)

        print("Testing has_task")
        self.assertTrue(self.cli.has_task('task-pancakes'))
        self.assertEqual(self.cli.has_task('task_pancakes'), False)

        print("Testing settings")
        self.cli.protect("off")
        off = self.cli.get_setting('watcher', 'protected')
        self.assertEqual(off, "off")
        self.cli.set_setting('watcher', 'protected', 'on')
        on = self.cli.get_setting('watcher', 'protected')
        self.assertEqual(on, "on")

        print('Testing get_section')
        section = self.cli.get_section('watcher')
        self.assertEqual(section.name, 'watcher')
        section = self.cli.get_section('task-pancakes')
        self.assertEqual(section.name, 'task-pancakes')
        section = self.cli.get_section('doesntexist')
        self.assertEqual(section, None)

        print('Testing get_task')
        task = self.cli.get_task('pancakes')
        self.assertEqual(task, None)
        task = self.cli.get_task('task-pancakes')
        tasks = self.cli.get_tasks()
        self.assertTrue(isinstance(tasks, list))
        self.assertEqual(len(tasks), 1)

        print('Testing Task')
        self.assertEqual(task.name, 'task-pancakes')
        params = task.export_params()
        print(params)
        correct = {'active': 'true',
                   'type': 'urls',
                   'url': 'https://www.google.com'}
        for key, value in correct.items():
            self.assertTrue(key in params)
            self.assertEqual(value, params[key])
        self.assertTrue(task.valid)
        self.assertEqual(task.type, 'urls')

        print("Testing inspect and list")
        self.cli.inspect()
        watchers = self.cli.list()
        self.assertTrue('pancakes' in watchers)
        self.assertTrue(isinstance(watchers, list))

    def test_schedule(self):
        self.assertEqual(self.cli.has_schedule(), False)


if __name__ == '__main__':
    unittest.main()



from .terminal import (
    run_command,
    get_installdir,
    get_watchme_env,
    which
)
from .fileio import (
    copyfile,
    get_userhome,
    get_user,
    get_host,
    generate_temporary_file,
    mkdir_p,
    get_tmpdir,
    print_json,
    read_file,
    read_json,
    write_file,
    write_json
)



'''

Copyright (C) 2019 Vanessa Sochat.

This Source Code Form is subject to the terms of the
Mozilla Public License, v. 2.0. If a copy of the MPL was not distributed
with this file, You can obtain one at http://mozilla.org/MPL/2.0/.

'''

from watchme import get_watcher
from watchme.logger import bot

def main(args, extra):
    '''edit the configuration for a watcher task
    '''    
    # Required - will print help if not provided
    name = args.watcher[0]
    action = args.action[0]
    task = args.task[0]

    # Get the watcher (exits if doesn't exist)
    watcher = get_watcher(name, base=args.base) 

    # Exit if the user doesn't provide a time
    if extra is None:
        bot.exit('Please provide one or more items to %s' % action)

    key = extra[0]
    value = None
    if action in ['add', 'update']:
        if len(extra) != 2:
            bot.exit('You must do watchme <watcher> add <key> <value>')                 
        value = extra[1]

    # Ensure the task exists        
    watcher.edit_task(task, action, key, value)



'''

Copyright (C) 2019 Vanessa Sochat.

This Source Code Form is subject to the terms of the
Mozilla Public License, v. 2.0. If a copy of the MPL was not distributed
with this file, You can obtain one at http://mozilla.org/MPL/2.0/.

'''

from watchme.command.create import ( 
    create_watcher_base, 
    create_watcher
)


def main(args, extra):
    '''init means creating the watcher base, and the first (default) watcher.
    '''    

    # Create the base
    create_watcher_base(args.watcher, args.base)

    # Create the first default watcher
    if not args.create_empty:
        create_watcher(args.watcher)

    
